
#
# <20220307.1500.43>
#
##################### Start of Documentation  ###################################
# Commands:
# ::pwLocker::getPw       Asks for password given Id
# ::pwLocker::putPw       Encripts a password and puts it in the table
# ::pwLocker::patternGet  Returns entries that match a pattern
# ::pwLocker::delete      Deletes one or more entries
# ::pwLocker::init        Sets addresses of en/decript routine,
#                         table address & table save routine

# This is an attempt to provide a password locker to
# keep multiple passwords and recall them by id.
# Most of the time the id will be a VFS address and user.
# However we will not restrict the id so almost anything
# will work. Well, not really. We create password entries
# that start with "protocol" "://" "username" "@" "address"
# However, while we keep all of this, we allow "nil" protocol,
# and address. We replace "nil" address with "localhost" and
# assume that is also kept in the table.
# On fetching a password we decrypt it and pass it back as a result.
#
#
# The saved password then has three parts:
# 1) The ID, must be unique.
# 2) An encrypted password. (This may be 3 parts as well.)
# 3) Part-3, an optional 'list' containing things that may be useful, e.g.
#    on MSW it contains info on 'CIFS' entries that indicate what drive letter
#    to use and how to set up 'Display Names'
#
# This code does NOT encrypt or decrypt but rather calls
# other code to do that. It is hoped that at some time
# we may have more advanced encrypt/decrypt code.
#
# For password storage we use a provided list of (triplets).
# passed in by name.

# The three parts of the password are:
# a) the password
# b) the identfile
# c) the passphrase
#
# For the most part (always?) only ssh and friends use parts b) and c).
# ssh and friends include ssh, scp, sftp rsync and putty

# To initialize, call:
#
# ::pwLocker::init passwdList encryptRoutine decryptRoutine Save
#
#                  where passwdList is the name of the password list
#                  encryptRoutine and decryptRoutine are commands and
#                  possible leading parms. Save is an optional command to
#                  save a modified "passwdList" and is called whenever the
#                  list is modified (i.e. a new password is saved).
#                  The password of interest will be added to these as 
#                  they are called.
#
#                  The passwordList entry should be an Id (which is an arbitrary
#                  string) followed by the encrypted password followed by an
#                  arbitrary entries. If a password entry is replaced, these
#                  will be saved. As a matter of convience the port (i.e. :<num>)
#                  is only considered part of the id if so requested (see below).
#
#
# To get a password:
# ::pwLocker::getPw Id args
#
#                        where Id is the id the password is stored under.
#                        The Id is assumed to have URL format and may or may not
#                        contain the user name. Examples:
#                        ftp://user@address      a full url
#                        ftp://@address  (or ftp:://address)  has a null user name
#                        ://user@host   a null protocol used for user
#                        ://            null protocol, user &  host
#                                       (host taken as localhost)
#                        user           All null except user name, taken to mean
#                                       ://user@localhost or, more concisely
#                                       user@localhost
#                        ftp://*@address Any user name (glob characters may be used)
#                        Null protocol assumes a non-protocol related request such as
#                        might be used with "sudo".
#                        Id may be a pattern which is used to search for entries. This
#                        may result in more than one (or even all) entries to be
#                        returned as a list.

#                        If the user name is supplied and the id is found
#                        the password is returned. Otherwise, if "noprompt" is false
#                        a GUI is presented to ask for the password. If the username
#                        is changed as part of that response and no password is
#                        provided, the URL with the name modified is looked up
#                        (as if a new call). If a password is entered in the GUI then
#                        it is returned (even if the username is changed). In this
#                        case the caller has no clue that the name was changed.
#                       
#                        If "noprompt" is true -1 is returned on failure.
#                        If "returnall" is true, sucess returns the password and
#                        failure will return all fields from the GUI responce.

#                        If found and the "use pattern" fails to find
#                        a satisfying proc, an error is thrown.

# Args is a (possibly empty) set of doublets with the following meaning:

#    key:     value:
#    newpassword <bool>  If true an entry window will be displayed. In general if
#                        this is true the only other option should be 'prompt'
#                        (which is optional). The prompt window will be dynamic
#                        (only asking for three part passwords for ssh and sftp
#                        entries). 
#    user   <username>   If the Id has a null name, this is used. This is also
#                        be part of the requested input in the GUI.
#    newuser <bool>      If true, a new user name is allowed. If neither the Id
#                        or the "user" (above) is provided, this is defaulted to
#                        true, otherwise the default is false.
#    abort   <bool> or string  If string, throw error "<string>" if Cancel or window
#                        destroyed.
#                        If 1 throw "User aborted request" on Cancel or window
#                        destroyed.
#                        If 0 (default) return -1 on Cancel of window destroyed.
#    full     <bool>     If true, return all 3 parts of the password
#                        {password identfile passphrase} default is to just
#                        return the password part. This also affects the
#                        GUI prompt if no password is found. If true all three
#                        parts are requested, if false, only the password part.
#                        Default is false.
#    prompt   {info}     If prompting for a password, include {info} in the request.
#                        If 'prompt' is present and not {}, a prompt is forced even
#                        if a password is present. This should be the second call
#                        once it is found that the first password has failed.
#    startwith <list>    If prompting, use these <list> hints. <list> is taken as
#                        the three parts of a password. This would usually be used
#                        where a prior request failed to "open" the resource.
#                        If this option is coded, pwLocker goes immediately to
#                        the GUI prompt without looking for an existing entry in
#                        the locker.
#    useport   <bool>    Usually the port is excluded from the id search. If this
#                        is true, the port will be included. In any case, if
#                        port is in the id it will be included in any saved
#                        password id. The port string is ":<n>" where <n> is
#                        a decimal number of at least 1 digit.
#    noprompt <binary>   Do not prompt if no password found, just return -1
#    nocase   <binary>   It true, do a 'nocase' search.
#    display  <ops>      <ops> should be a dictionary with name followed by the
#                        starting value. Each entry will result in a display of 
#                        with the value given. E.G. display {z: {} newname foo}
#                        would give two entrys "z:" <blank> and "newname" foo.
#                        The returnall option will return these values.
#    returnall <bool>    If true, and a prompt was used (i.e. a GUI was presented)
#                        all the detail will be returned as a dict list
#                        e.g. password <value> op<n> <value> save <bool> name <value>
#                        In this case the password is saved or not depending on
#                        the presents of the "display" option and the "keep"
#                        option from the GUI result. If "keep" is true and "display"
#                        is not coded, the keep is honored. Otherwise, it is
#                        expected that the "display" results need to be examined
#                        and/or used by the caller to do the save.

#                        If returnall is true and the password exists all its
#                        details are returned as a dict. list with the following
#                        elements: id <full id i.e. type://name@address:port>,
#                        name <name>, type <type>, address <address> port <port>,
#                        password <full 3 part password> options <options list>

#  keys may be abbreviated (to any unique set of one or more letter(s)
#
#  Note that a -1 return is used for "no password" on a prompt. If 'returnall' is true
#  in this case, the returned password will be an empty string.
#  This is to allow a nil password as a possible return.
#  A -2 indicates a "Cancel" result.

# The following two routines are intended mainly to be used for maintance of the
# password locker.

# To get passwords using a pattern:
#
# ::pwLocker::patternGet pattern args
#
#                        The 'pattern' is applied to the ID.

# args are optional key value pairs as follows:
#
#  key       value        meaning
#  pattern2  pattern     This pattern will be applied to the entries part-3
#                        (the optional list). The whole of 'part-3' is treated as a
#                        string for this match.
#  case      <bool>      If true the search is done in a case sensitive manner
#                        (Usuall case is not important thus the default of nocase.)
#  !         0-3         Invert the pattern match sense:
#                        0 does nothing,
#                        1 invert pattern1 match
#                        2 invert pattern2 match (ignored if no pattern2)
#                        3 invert both patterns match

# The result of ::pwLocker::patternGet pattern args is a list of password entries.
# Each element of the list will be {Id password part-3}. Both 'password' and 'part-3'
# may be lists ('password' has three part, see above).

# To delete one or more entries:

# ::pwLocker::delete pattern args
#
# args are optional key value pairs the same as ::pwLocker::patternGet


# To put a password:
#
# ::pwLocker::putPw newId newPassWord args
#
#                        If 'newId' is already in the password list, that
#                        entry is deleted. The new entry is then added to
#                        the front of the list (after being encrypted)
#                        Here "args" is a list of arbitrary length
#                        which this code just appends these to the entry.

##################### End of Documentation  ###################################
#
# the debug flag turns off the initial clearing of the pwroutines
# and the security checks...
set pwLockerDebug 1
namespace eval pwLocker {
  # Set to 0 to allow reload for debugging
  if {$::pwLockerDebug} {
    variable pwAddress      
    variable encryptRoutine 
    variable decryptRoutine 
    variable saveFun      
  } else {
    variable pwAddress      {}
    variable encryptRoutine {}
    variable decryptRoutine {}
    variable saveFun        {}
  }
  
  proc init {name encrypt decrypt {SaveFun {}}} {
    variable encryptRoutine $encrypt
    variable decryptRoutine $decrypt
    variable pwAddress      $name
    variable saveFun        $SaveFun
  }
  
  proc getPw {id args} {
    variable {}
    variable pwAddress
    variable encryptRoutine 
    variable decryptRoutine
    array unset ::pwLocker:: *
    # was init called???
    foreach par {pwAddress encryptRoutine decryptRoutine} {
      if {[set $par] == {}} {
	return -code error\
	    "pwLocker::getPw detected no or incorrect pwLocker::init call."
      }
    }
    # Ok, lets take a look...
    set keys {full prompt startwith noprompt nocase newuser newpassword
      abort useport display returnall user}
    set binKeys {abort full useport noprompt nocase returnall newuser newpassword} 
    foreach key $keys {
      if {$key in $binKeys} {
	set $key 0
      } else {	
	set $key {}
      }
    }
    lappend abort "User aborted request." 0
    while {$args ni {{} {{}}} } {
      set args [lassign $args opt]
      frputs args opt
      # is this a list?
      if {[llength $opt] > 1} {
	# yes, break it up and push it ahead of the rest
	set args [linsert $args 0 {*}$opt]
	continue
      }
      # not a list, get value part...
      set args [lassign $args value]
      frputs value
      set tar [lsearch -glob -all -inline $keys "$opt*"]
      if {$tar == -1 || [llength $tar] != 1} {
	error "pwLocker::getPw:\
               option ($opt) must be one of: \"$keys\" but was \"$tar\""
      }
      set ${tar}Provided 1
      if {$tar in $binKeys} {
	set $tar $value
      } else {
	lappend $tar {*}$value
      }
    }
    if {$newpassword} {
      foreach key $keys {
	if {$key in $binKeys} {
	  set $key 0
	} else {	
	  set $key {}
	}
      }
      set newuser [set newpassword 1]
    }
    
    set (-full-) $full
    frputs full (-full-) prompt Id id
    if {![regexp {(([^:]+)://){0,1}(([^@]+)($|@($|([^/]+)($|/.*$))))} $id ->\
	      i1 protcol i3 uname i5 i6 addr  i8]} {
      # Lets allow a null id if a newuser is specified.
      if {!$newuser || $id != {}} {
	# Very bad news, not one of the id formats we recognize
	return -code error "Password Locker: unrecognized id format"
      }
    }
    # Break the Id apart and fill in the blanks, if possible. We assume we are
    # either dealing with a URL (protocol://{user@}add{/path}
    # Or a simple password      (user{@add}) which is the same as (://user{@add})
    #                           and we treat nil 'add' as 'localhost'
    # Note that in the simple password the address is optional
    # while in the URL the user is optional...
    #
    lassign $id VFSadd VFSpro VFSpath 
    if {[IsVFS $id] || [IsVFS -nil-$id] || [IsVFS -nil-://$id]} {
      append VFSpro "://"
    }     
    lassign [VFSaddrSplit $VFSadd] (user) VFSadd port
    # if only one of address & user, lets assume it is user
    # and local...
    if {$(user) == {} && $VFSadd != {}} {
      set (user) $VFSadd
      set VFSadd "localhost"
    }
    if {$(user) == {} && ![info exists newuserProvided]} {
      set newuser 1
    }

    set adds $VFSadd
    if {$useport != 1} {
      set adds [regsub {:\d+} $VFSadd {*}]
    }
       
    if {$user != {}} {
      # specified user overrides one in the id
      set (user) [lindex $user end]
    }
    while {1} {
      set Id [expr {$VFSpro != "-nil-" ? $VFSpro : ""}]
      append Id [expr {$(user) == {} ? {} : "$(user)@"}] $adds
      # we need to special case cifs because the VFS 'path' is really
      # part of the share
      frputs VFSpro VFSpath Id adds 
      if {[string match -nocase $VFSpro  "cifs://"]} {
	set end [expr {[set pend [string first / $VFSpath 1]] == -1 ?
		       "end" : $pend -1}]
	append Id / [string range $VFSpath 0 $end]
      }
      # Got the options, do the work
      lassign {} (password) (disk) (passphrase)
      if {[llength $prompt] > 0 || $newpassword || $startwith != {} } {
	set la {}
	if {$startwith != {}} {
	  lassign $startwith (password) (disk) (passphrase)
	}
      } else {
	set la [lsearch -glob -inline -all\
		    {*}[expr {$nocase ? "-nocase" : ""}]\
		    -index 0 [set $pwAddress] $Id]
	if {$VFSpro == "-nil-"} {
	  lappend la {*}[lsearch -glob -inline -all\
			     {*}[expr {$nocase ? "-nocase" : ""}]\
			     -index 0 [set $pwAddress] ://$Id]
	}
	frputs la Id VFSpro VFSpath end pend
      }
      frputs la prompt (user) Id
      set (fullid) $Id
      switch  -exact [llength $la] {
	0 {
	  # Password not found, here we ask for it with several options...
	  if {$noprompt == 1} {
	    return -1
	  }
	  set la2 {}
	  set pr {}
	  set co 1
	  set mesChoose {}
	  lappend pr  [list [_ "Full Id:"]\
			   [list -textvariable [namespace current]::(fullid)]]
	  if {$newuser} {
	    incr co
	    lappend pr  [list [_ "User name:"] \
			     [list -textvariable [namespace current]::(user) \
				  -state [expr {$newuser ? "normal" : "disabled"}]]]
	  }
	  set (-showCount-) $co
	  incr co
	  lappend pr  [list [_ "Password:"] \
			   [list -textvariable [namespace current]::(password)\
				-show "*" ]]
	  if {$(-full-) == 1} {
	    incr co 2
	    lappend pr [list [_ "ssh private key disk address:"]\
			    [list -textvariable [namespace current]::(disk) \
				 -show "*" ]]
	    lappend pr [list [_ "Pass phrase:"]\
			    [list -textvariable [namespace current]::(passphrase)\
				 -show "*" ]]
	  }
	  set opc 0
	  # We use the "-" to avoid collisions with our vars
	  frputs display
	  if {$display != {}} {
	    foreach {val op} $display {
	      incr co
	      set (${val}) $op
	      lappend pr [list $val:\
			      [list -textvariable [namespace current]::(${val})]]
	    }
	  }
	  #set (name) $user
	  set (-showpw-) 0
	  set (keep) 0
	  set (-user-) $(user)
	  set (-pw-) [list $(password) $(disk) $(passphrase)]
	  incr co 4
	  set dId $VFSpro
	  append dId [expr {$(user) == {} ? $VFSadd : "$(user)@$VFSadd"}]
	  frputs dId (user) VFSadd
	  set (-uni-) [incr ::uni]
	  set (VFSpro) $VFSpro
	  set (VFSpath) $VFSpath
	  set (VFSadd) $VFSadd
	  trace add variable [namespace current]::(user) write ::pwLocker::wrUser
	  trace add variable [namespace current]::(fullid) write ::pwLocker::wrId
	  set rt [smart_dialog .password_entry_dialog$(-uni-) . PassWord \
		      [concat [list $mesChoose {}  [_ "Enter password for %s\n\
                             \nOK activates, cancel or window-delete cancels."\
							$dId] "\n"] $prompt]\
		      0 $co  [list {*}$pr \
				  [list [_ "Keep Password"] \
				       [list -variable [namespace current]::(keep)]]\
				  [list [_ "Show password"] \
				       [list -variable [namespace current]::(-showpw-)\
					    -command  [namespace current]::PwShow]]\
				  [list [_ "OK"]]\
				  [list [_ "Cancel"]]\
				  ]\
		      ]
	  frputs "[namespace current] " rt co abort
	  if {$rt == ($co -1)} {set rt -2}
	  if {$rt < 0} {
	    if {[string is boolean [set v [lindex $abort end]]]} {
	      frputs v
	      if {!$v} {
		return $rt
	      } else {
		return -code error [lindex $abort 0]
	      }
	    }
       	    return -code error "$v"
	  }
	  set pw [list $(password) $(disk) $(passphrase)]
	  # If the username was changed....
	  # if the pass word was not changed, check for the new user
	  frputs (user) (-user-) (-pw-) pw
	  # if {$(user) != $(-user-) && $(-pw-) == $pw} {continue}

	  # We have a new user and a new password.
	  # append (id) $VFSpro
	  # if {$(user) != {}} {
	  #   append (id) $(user) "@"
	  # }
	  # append (id) $VFSadd / $VFSpath
	  set (id) $(fullid)
	  frputs (id) pw
	  if {$(keep) && $display == {}} {
	    putPw $(id) $pw   
	  }
	  if {$returnall} {
	    frputs
	    # build the id based on the new info
	    return [array get ::pwLocker:: {[a-zA-Z]*}]
	  }
	  if {$(-full-) != 1} {
	    return $(password)
	  }
	  return $pw
	}
	default {
	  # We found a password
	  # if {[info level] <= 1 && !$::pwLockerDebug} {
	  #   array set info [info frame -2]
	  #   if {![string match *MenuInvoke $info(proc)]} {
	  #     frputs info(proc)
	  #     return -code error\
	  # 	  "Illegal attempt to access passwordLocker for $id"
	  #   }
	  # }
	  array unset info
	  # We may have more than one. This is Ok only if the resulting
	  # passwords are all the same.
	  
	  set pwds {}
	  set pwdCount 0
	  foreach ent $la {	  
	    set fpw [eval [list {*}$decryptRoutine [lindex $ent 1]]]
	    if {$fpw ni $pwds} {
	      lappend pwds $fpw
	      incr pwdCount
	    }
	  }
	  frputs pwds pwdCount la
	  if {$pwdCount > 1} {
	    return -code error "Password locker found $pwdCount passwords \
                   for\n$id"
	  }
	  if {!$returnall} {
	    if {$(-full-) == 1} {
	      return $fpw
	    }
	    return [lindex $fpw 0 0]
	  }
	  # build the dict. list. entries are:
	  #   id <full id i.e. type://name@address:port>,
	  #   name <name>,
	  #   type <type>,
	  #   address <address> (IsVFS combines user add & port here)
	  #   port <port>,
	  #   password <full 3 part password>
	  #   options <options list>

	  
	  lappend result id [lindex $ent 0] password $fpw options [lindex $ent 2]
	  if {![IsVFS $id]} {
	    # This should never happen, but just in case...
	    lappend add {} port {} type {} name{}
	  } else {
	    lappend result type $VFSpro
	    # Make sure we allow "@" in the name...
	    if {[set namEnd [string last "@" $VFSadd]] != -1} {
	      lappend result name [string range $VFSadd 0 $namEnd-1]
	    } else {
	      set namEnd "end"
	    }
	    lassign [split [string range $VFSadd $namEnd+1 end] :] add port
	    lappend result address $add port $port
	  }
	  return $result	    
	}
      }
    }
  }

  proc parseAndGetMatches {pattern ops} {
    variable pwAddress
    lassign {{} 0 0 0} pattern2 case -nocase !
    set legal {pattern2 case -nocase !}

    if {[llength $ops] & 1} {
      return -code error "Argument list must have an even number of items"
    }
    foreach {op value} $ops {
      set thisOp [lsearch -inline $legal ${op}*]
      if {$thisOp == {}} {
	return -code error "Bad option found ($op). Must be one of '$legal'."
      }
      set $thisOp $value
    }
    if {![string is boolean $case]} {
      return -code error "Found '$nocase' for 'nocase option' must be boolean."
    }
    if {${!} ni {0 1 2 3}} {
      return -code error "Found '${!}' for '!'. Must be '0, 1 ,2 or 3."
    }
    set case [expr {${-nocase} || !$case ? "-nocase" : ""}]
    set not1 [expr {${!} & 1 ? "-not" : ""}]
    set result {}
    set result [lsearch -all {*}$not1 {*}$case -index 0 [set $pwAddress] $pattern]
    if {$result == {} || $pattern2 == {}} {return $result}
    # We have a pattern2, run the result through that test

    set result2 {}
    foreach ent $result {
      if {[string match {*}$case $pattern2 [lindex [set $pwAddress] $ent 2]] ^ (${!} & 2)} {
	lappend result2 $ent
      }
    }
    return $result2
  }

  proc patternGet {pattern args} {
    variable pwAddress
    variable decryptRoutine
    set indexs [parseAndGetMatches $pattern $args]
    set rtn {}
    foreach ent $indexs {
      lassign [lindex [set $pwAddress] $ent] id pw option
      # frputs id pw option
      set fpw [eval [list {*}$decryptRoutine $pw]]
      lappend rtn [list $id $fpw $option]
    }
    return $rtn
  }

  # The pattern delete. This might be coded more efficiently but we want simplicity
  # not speed. We reverse the list of indices's so the following indices's still work.
  
  proc delete {pattern args} {
    variable pwAddress
    variable saveFun
    set count 0
    foreach index [lreverse [parseAndGetMatches $pattern $args]] {
      set $pwAddress [lreplace [set $pwAddress] $index $index]
      incr count
    }
    if {$saveFun != {}} {
      eval [list {*}$saveFun]
    }
    
    return $count
  }
  
  # These routines service the dialog box
  proc PwShow {} {
    variable {}
    set add $(-showCount-)
    frputs add
    set showChar [expr {$(-showpw-) ? {} : {*}}]
    .password_entry_dialog$(-uni-).$add config -show $showChar
    if {$(-full-)} {
 
      .password_entry_dialog$(-uni-).[incr add] config -show $showChar
      .password_entry_dialog$(-uni-).[incr add] config -show $showChar
    }
  }

  # Write to user... 
  proc wrUser {args} {
    variable {}
    append nid $(VFSpro) $(user)\
	[expr {$(user) == {} ? "" : "@"}]\
	$(VFSadd) [expr {$(VFSpath) == {} ? "" : "/$(VFSpath)"}]
    trace remove variable [namespace current]::(fullid) write ::pwLocker::wrId
    set (fullid) $nid
    trace add variable [namespace current]::(fullid) write ::pwLocker::wrId
    
  }
  proc wrId {args} {
    variable {}
    set end [string last "@" $(fullid)]
    if {$end > 0} {
      trace remove variable [namespace current]::(user) write ::pwLocker::wrUser
      set (user) [string range $(fullid) [string last "/" $(fullid) $end]+1 $end-1]
      trace add variable [namespace current]::(user) write ::pwLocker::wrUser
    }
  }

  
  # args will be appended to the password entry after the password
  # for vfs and cifs passwords these are drive and flag words
  
  proc putPw {newId newPw args} {
    variable pwAddress
    variable encryptRoutine
    variable saveFun
    set newId [regsub -all {\*} $newId {}]
    set extra {}
    if {[llength $args] != 0 && [lindex $args 0] != {}} {
      set extra $args
    }
    frputs args [llength $args]
    while {[set la [lsearch -exact -index 0 [set $pwAddress] $newId]] != -1} {
      if {[llength $args] == 0} {
	set extra [lindex [set $pwAddress] $la 2]
      }
      set $pwAddress [lreplace [set $pwAddress] $la $la]
    }
    set pw [eval [list {*}$encryptRoutine $newPw]]
    # we want the actual value, not the list protected value
    # set extra [list [lindex $extra 0] [lindex $extra 1]]
    
    frputs extra args ">[string trim $extra]< "
    set $pwAddress [linsert [set $pwAddress] 0 [list $newId $pw $extra]]
    if {$saveFun != {}} {
      eval [list {*}$saveFun]
    }
  }
 
}
